class Widget():
    """ Base class for all Widgets.

    Attributes:
        id (int): The ID of the Widget. Each created Widget have his own ID.
        jquery_fn (str): The name of the JQuery function to use to get the value
            of the widget from the HTML page.
        cls (list of str): List of CSS classes to apply to this widget.
    """

    _id = 0             # Counter for the next Widget ID
    _declared = []      # List of all declared Widget

    def __init__(self, cls=None):
        """ Constructor. """
        Widget._id += 1
        Widget._declared.append(self)
        self.id = Widget._id
        self.jquery_fn = "text"

        if cls is None:
            self.cls = []
        elif isinstance(cls, str):
            self.cls = [cls]
        elif isinstance(cls, list):
            self.cls = [c for c in set(cls) if isinstance(c, str)]
        else:
            raise ValueError("The given CLS class should be a string or a list of string")

    def add_css_class(self, attr):
        """ Utils method to add the class attribute in the given dictionary.
        This dictionary can then be used in `dominate`.

        Arguments:
            attr (dict): The attributes dictionary to be used in `dominate` tag
                class.
        """
        if len(self.cls) != 0:
            attr["cls"] = " ".join(self.cls)

    def html(self):
        """ Method to get the `dominate` HTML of the widget. This HTML needs to
        be rendered.

        Returns:
            dominate.document: HTML document corresponding to the widget.
        """
        raise NotImplementedError()

    def ajax(self):
        """ Method to get the Ajax request (if any) of the widget. Or `None` if
        there is no Ajax call for this widget.

        Returns:
            Ajax: Ajax request. `None` if there is no Ajax request.
        """
        return None

    def get(self):
        """ Method to get the current value of the widget.

        Returns:
            object: Current value of the widget.
        """
        raise NotImplementedError()

    def get_str(self):
        """ Method to get the current value of the widget, as a string.

        Returns:
            str: Current value of the widget.
        """
        return str(self.get())

    def set(self, x):
        """ Method to set the current value of the widget.

        Arguments:
            x (str): Value of the widget to set.
        """
        raise NotImplementedError()



def route_to_filename(route):
    """ Function to transform a route to a filename.

    Arguments:
        route (`str`): Route name to transform to a filename.

    Returns:
        `str`: Filename corresponding to the route.
    """
    return route.replace('/', '_')



from dominate.tags import input_

from swole.widgets.base import Widget


class Input(Widget):
    """ Widget to create an input.

    Attributes:
        type (`str`): Type of input.
        placeholder (`str`): Placeholder of input.
    """
    def __init__(self, type="number", placeholder=None, **kwargs):
        """ Constructor.

        Arguments:
            type (`str`, optional): Type of the input. Defaults to `number`.
            placeholder (`str`, placeholder): Placeholder for the input. If
                `None`, no placeholder is used. Defaults to `None`.
        """
        super().__init__(**kwargs)
        self.type = type
        self.placeholder = placeholder
        self.value = placeholder
        self.jquery_fn = "val"

    def html(self):
        attributes = {
            "id": self.id,
            "type": self.type,
        }
        self.add_css_class(attributes)

        if self.placeholder is not None:
            attributes["placeholder"] = self.placeholder

        return input_(**attributes)

    def get(self):
        return self.value

    def set(self, x):
        if self.type == "number" and isinstance(x, str):
            try:
                # Find if it's an int or a float
                number = float(x)
                integer = int(number)
                # number, integer = float(x), int(x)
                self.value = integer if number == integer else number
            except ValueError:      # Empty input
                self.value = 0
        else:
            self.value = x



from dominate.tags import p

from swole.widgets.base import Widget


class Markdown(Widget):
    """ A general widget to write Markdown.

    Attributes:
        content (`str`): Markdown content.
    """
    def __init__(self, content="", **kwargs):
        """ Constructor.

        Arguments:
            content (`str`, optional): Markdown content. Defaults to ``.
        """
        super().__init__(**kwargs)
        self.content = content

    def html(self):
        # TODO : Full markdown parsing + convertion to dominate tags
        return p(self.content, id=self.id)

    def get(self):
        return self.content

    def set(self, x):
        self.content = x



from dominate.tags import button

from swole.widgets.base import Widget


class Button(Widget):
    """ Widget to create a button.

    Attributes:
        text (`str`): Text of the button.
        primary (`bool`): Wether this button is primary or not.
    """
    def __init__(self, text="Button", primary=False, onclick=None, **kwargs):
        """ Constructor.

        Arguments:
            text (`str`, optional): Text of the button. Defaults to `Button`.
            primary (`bool`, optional): Wether this button is primary or not.
                Defaults to `False`.
            onclick (`Ajax`, optional): Ajax request to call if the button is
                clicked. If `None`, nothing happend on click. Defaults to
                `None`.
        """
        super().__init__(**kwargs)
        self.text = text
        self.primary = primary
        self.onclick = onclick

        if primary:
            self.cls.insert(0, "button-primary")

    def html(self):
        attributes = {"id": self.id}
        self.add_css_class(attributes)

        if self.onclick is not None:
            attributes["onclick"] = "callback_{}()".format(self.onclick.id)

        return button(self.text, **attributes)

    def ajax(self):
        return self.onclick if self.onclick is not None else None

    def get(self):
        return self.text

    def set(self, x):
        self.text = x



import os
import re


class Skin():
    """ Class representing the skin to use for the page. A skin is basically a
    CSS file, with additional imports for CSS libraries and fonts.

    Attributes:
        path (str): Path of the CSS file for this skin.
        libs (list of str): External CSS libraries to import additionally.
        fonts (list of str): External fonts to import additionally.
    """

    def __init__(self, name="base", path=None):
        """ Constructor. Try to locate the file, then read it to extract
        external CSS resources links.

        Arguments:
            name (str, optional): The name of the skin to use. Defaults to
                `base`.
            path (str, optional): If different than `None`, the path of the CSS
                file to use. If different than `None`, `name` argument is
                ignored. Useful to provide custom skin. Defaults to `None`.
        """
        # Get the file's path
        if path is None:
            # Get the path of the skin from its name
            skin_dir = os.path.dirname(os.path.abspath(__file__))
            path = os.path.join(skin_dir, "{}.css".format(name))
        self.path = path

        self.libs, self.fonts, self.rules = self._extract_ext_css(self.path)

    def _extract_ext_css(self, path):
        """ Helper method to extract external links from a skin file.

        Arguments:
            path (str): The path of the skin file.

        Returns:
            list of str: List of external CSS libraries links.
            list of str: List of external Fonts links.
            str: The custom CSS rules for this skin.
        """
        with open(path) as f:
            content = f.read()

        pattern = re.compile(r"/\*.*?\*/", re.DOTALL)
        uncommented_content = re.sub(pattern, "", content).strip()

        lines = content.split('\n\n')
        if len(lines) < 2:
            return [], [], uncommented_content
        libs_content, fonts_content, *_ = lines

        libs_links = self._extract_links(libs_content)
        font_links = self._extract_links(fonts_content)
        return libs_links, font_links, uncommented_content

    def _extract_links(self, content):
        """ Helper function to extract links from a string representing several
        links commented out in the CSS file.

        Arguments:
            content (str): String to clean and extract links.

        Returns:
            list of str: Cleaned list of links.
        """
        # Each line is a link
        lines = content.split('\n')

        # Link are in the comments : filter out non-comment
        links = [line for line in lines if line.startswith("/* ") and line.endswith(" */") and len(line) > 6]

        return [link[3:-3] for link in links]



import os
from typing import Dict

from fastapi import FastAPI
from starlette.responses import FileResponse
import uvicorn

from swole.core.page import Page, HOME_ROUTE
from swole.core.utils import route_to_filename
from swole.widgets import Widget


SWOLE_CACHE = "~/.cache/swole"


class Application():
    """ Class representing an application. An application is the englobing,
    object, which list all possible routes.

    Attributes:
        pages (`dict`): Dictionary[`str`: `Page`] listing all possible routes
            and their corresponding Page.
        files (`dict`): Dictionary[`str`: `str`] listing all possible routes and
            their corresponding saved HTML file. This attribute is set only
            after calling the method `write()`.
        fapi (`fastapi.FastAPI`): FastAPI app.
    """
    def __init__(self, pages=None):
        """ Constructor.

        Arguments:
            pages (`list`, optional): List of `Page`. Defaults to `None`.
        """
        self.pages = {p.route: p for p in pages} if pages is not None else {}
        self.files = None
        self.fapi = FastAPI()

    def add(self, pages):
        """ Method to add pages to the application.

        Arguments:
            pages (`Page` or `list`): Page or list of Page to add.
        """
        if isinstance(pages, list):
            for page in pages:
                self._add(page)
        else:
            self._add(pages)

    def _add(self, page):
        if not isinstance(page, Page):
            raise ValueError("Expected a Page type, got a {} type instead".format(type(page)))

        if page.route in self.pages:
            raise ValueError("This route ({}) is already set".format(page.route))

        self.pages[page.route] = page

    def assign_orphan_widgets(self):
        """ Method finding orphan widgets if any, and assigning it to the Home
        page if there is no Home page already. This allow a very simple and easy
        way to use the library.
        """
        if HOME_ROUTE in self.pages:
            # Home page already defined, nothing to do
            return

        home = Page()
        assigned_widgets = set().union(*[page.widgets for page in self.pages.values()])
        for w in Widget._declared:
            if w not in assigned_widgets:       # Orphan !
                home.add(w)

        self.pages[HOME_ROUTE] = home

    def write(self, folder=SWOLE_CACHE):
        """ Method to write the HTML of the application to files, in order to
        later serve it.

        Arguments:
            folder (`str`, optional): Folder where to save HTML files. Defaults
                to SWOLE_CACHE.
        """
        os.makedirs(folder, exist_ok=True)

        self.files = {}         # Route -> HTML file
        self.callbacks = {}     # Callback ID -> (Page, Ajax)
        for route, page in self.pages.items():
            # Write HTML of the page
            html_str = page.html().render()
            path = os.path.join(folder, "{}.html".format(route_to_filename(route)))
            with open(path, 'w') as f:
                f.write(html_str)
            self.files[route] = path

            # Save also callbacks (along with their page)
            for aj in page.ajax():
                self.callbacks[str(aj.id)] = (page, aj)

    def define_routes(self):
        """ Method defining the routes in the FastAPI app, to display the right
        HTML file.
        """
        # Define the pages' routes
        for route, html_file in self.files.items():
            @self.fapi.get(route)
            def index():
                return FileResponse(html_file)

        # Define the callback route
        @self.fapi.post("/callback/{callback_id}")
        def callback(callback_id, inputs: Dict[str, str]):
            page, ajax = self.callbacks[callback_id]
            return ajax(page, inputs)

    def serve(self, folder=SWOLE_CACHE, host='127.0.0.1', port=8000, log_level='info'):
        """ Method to fire up the FastAPI server !

        Arguments:
            folder (`str`, optional): Folder where to save HTML files. Defaults
                to SWOLE_CACHE.
            host (`str`, optional): Run FastAPI on this host. Defaults to
                `127.0.0.1`.
            port (`int`, optional): Run FastAPI on this port. Defaults to
                `8000`.
            log_level (`str`, optional): Log level to use for FastAPI. Can be
                [`critical`, `error`, `warning`, `info`, `debug`, `trace`].
                Defaults to `info`.
        """
        self.assign_orphan_widgets()
        self.write(folder=folder)
        self.define_routes()

        uvicorn.run(self.fapi, host=host, port=port, log_level=log_level)



def ajax(*args):
    """ Decorator to define Ajax objects easily.

    Arguments:
        args (list of Widget): Widget to use as input.

    Returns:
        Ajax: Ajax object with the given callable and inputs.
    """
    def define_ajax(f):
        return Ajax(callback=f, inputs=args)
    return define_ajax


class Ajax():
    """ Class representing an AJAX request. It is used as callback to update
    the webpage dynamically.

    Attributes:
        callback (callable): Function to execute when the AJAX is called.
        inputs (list of Widget): Inputs to the callback.
        trigger (Widget): Widget triggering the AJAX request.
    """

    _id = 0

    def __init__(self, callback, inputs):
        """ Constructor. Make the AJAX from the function to execute, and define
        the inputs as given.

        Arguments:
            callback (callable): Function to execute when the AJAX is called.
            inputs (list of Widget): Inputs to the callback.
            trigger (Widget): Widget triggering the AJAX request.
        """
        Ajax._id += 1
        self.id = Ajax._id
        self.callback = callback
        self.inputs = inputs

    def __call__(self, page, input_data):
        """ Main method, being called by the application with the right inputs.
        This method keep track of the value of each widget of the page, and
        based on what was changed, return only the element to change in the
        page.

        Arguments:
            page (Page): Page calling the AJAX.
            input_data (dict): Inputs data retrieved from the page after the
                AJAX request was triggered. It's a dict of `str` -> `str` where
                the key is the ID of the widget and the value is the value of
                the widget.

        Returns:
            dict: Dictionary of ID to value, containing all widgets to update.
        """
        # Update value that are given
        for inp in self.inputs:
            inp.set(input_data[str(inp.id)])

        # Keep dictionary of all current value
        prev_values = {str(w.id): w.get_str() for w in page.widgets}

        # Call the callback with the right inputs
        self.callback(*[inp.get() for inp in self.inputs])

        # After callback, see what changed
        now_values = {str(w.id): w.get_str() for w in page.widgets}

        to_update = {}
        for k in prev_values:
            if prev_values[k] != now_values[k]:
                to_update[k] = now_values[k]
        return to_update

    def js(self):
        """ Method writing Javascript equivalent for this AJAX request.

        Returns:
            `str`: Javascript equivalent.
        """
        return """
      function callback_{0}() {{
          var data = {{{1}}};
          callback({0}, data);
      }}""".format(self.id, ",".join(['{0}: $("#{0}").{1}()'.format(i.id, i.jquery_fn) for i in self.inputs]))



import setuptools

with open("README.md", "r", encoding='utf-8') as fh:
    long_description = fh.read()

with open('requirements.txt') as fr:
    reqs = fr.read().strip().split('\n')


setuptools.setup(
    name="swole",
    version="0.0.2",
    author="Nicolas REMOND",
    author_email="remondnicola@gmail.com",
    description="Streamlit, but better.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/astariul/swole",
    packages=setuptools.find_packages(),
    package_data={'swole': ['skins/*.css']},
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=reqs,
)



import dominate
from dominate.tags import script, link, style, div, br
from dominate.util import raw

from swole.widgets.base import Widget
from swole.skins import Skin


HOME_ROUTE = "/"
JQUERY_CDN = "https://ajax.googleapis.com/ajax/libs/jquery/3.5.1/jquery.min.js"
COMMON_JS = """
      function callback(id, data) {
          $.ajax({
              type: "POST",
              url: `/callback/${id}`,
              data: JSON.stringify(data),
              success: function (data) {
                  for (var key in data) {
                      $(`#${key}`).val(data[key]);
                      $(`#${key}`).text(data[key]);
                  }
              },
              failure: function (errMsg) {
                  alert(errMsg);
              }
          });
      }
    """


class Page():
    """ Class representing a page.

    Attributes:
        route (`str`): The route to access this page.
        skin (Skin): The Skin object for this page.
        title (`str`): The title of the page.
    """
    def __init__(self, route=HOME_ROUTE, skin="base", skin_path=None, title="Home"):
        """ Constructor.

        Arguments:
            route (`str`, optional): The route to access this page. Defaults to
                `/`.
            skin (`str`, optional): The name of the skin to use for this page.
                If `None` is given, no skin is loaded. Defaults to `base`.
            skin_path (`str`, optional): The path of the Skin file to use. If
                different than `None`, the `skin` argument is ignored, and this
                file is used instead. Useful to provide custom Skin file.
                Defaults to `None`.
            title (`str`, optional): The title of the page. Defaults to `Home`.
        """
        self.route = route
        self.skin = Skin(name=skin, path=skin_path) if skin is not None else None
        self.title = title
        self.widgets = []

    def html(self):
        """ Method to get the `dominate` HTML of the page. This HTML needs to be
        rendered.

        Returns:
            dominate.document: HTML document corresponding to the page.
        """
        doc = dominate.document(title=self.title)

        # Add external files (Skin)
        if self.skin is not None:
            with doc.head:
                for ref in self.skin.libs:      # Libs
                    link(rel='stylesheet', crossorigin='anonymous', href=ref)

                for ref in self.skin.fonts:      # Fonts
                    link(rel='stylesheet', type='text/css', href=ref)

                if self.skin.rules != "":
                    style(raw(self.skin.rules))

        # Add Widgets HTML to the page
        main_div = div(cls="container")
        for w in self.widgets:
            main_div.add(w.html())
            main_div.add(br())
        doc.add(main_div)

        # Add Javascript code to the page
        js_str = "\n\n".join([a.js() for a in self.ajax()])
        if js_str != '':
            doc.add(script(src=JQUERY_CDN))
            doc.add(script(raw(js_str + "\n\n" + COMMON_JS)))

        return doc

    def add(self, widget):
        """ Method to add a widget to this page.

        Arguments:
            widget (`Widget`): Widget to add.
        """
        if not isinstance(widget, Widget):
            raise ValueError("The given argument is not a widget : {}".format(widget))

        self.widgets.append(widget)

    def ajax(self):
        """ Method to retrieve all Ajax requests of this page.

        Returns:
            list of Ajax: List of Ajax requests.
        """
        ajax = [w.ajax() for w in self.widgets]
        return [a for a in ajax if a is not None]

    def __enter__(self):
        """ Context manager for easy definition of Widgets inside the page :
        Remember the declared widgets at this point.
        """
        self._from = len(Widget._declared)
        return self

    def __exit__(self, type, value, traceback):
        """ Context manager for easy definition of Widgets inside the page :
        Add any newly declared widgets.
        """
        new_widgets = Widget._declared[self._from:]
        for w in new_widgets:
            self.add(w)
